'''
Client for REST APIs.
'''

from __future__ import print_function, division

import collections
import re
from syndicate import data as m_data
from syndicate.adapters import sync as m_sync, async as m_async
from tornado import concurrent


class ServiceError(Exception):
    pass


class ResponseError(ServiceError):

    def __init__(self, response):
        self.response = response

    def __str__(self):
        return '%s(%s)' % (type(self).__name__, self.response)


class Service(object):
    """ A stateful connection to a service. """

    default_page_size = 100
    urlpartition = re.compile('([?;])')

    @staticmethod
    def default_data_getter(response):
        if response.error:
            raise response.error
        content = response.content
        if not content:
            return None
        if not content['success']:
            raise ResponseError(content)
        return content.get('data')

    @staticmethod
    def default_meta_getter(response):
        content = response.content
        if not content:
            return None
        return content.get('meta')

    def __init__(self, uri=None, urn='', auth=None, serializer='json',
                 data_getter=None, meta_getter=None, trailing_slash=True,
                 async=False, adapter=None, request_timeout=None):
        if not uri:
            raise TypeError("Required: uri")
        self.async = async
        self.auth = auth
        self.filters = []
        self.trailing_slash = trailing_slash
        self.uri = uri
        self.urn = urn
        self.request_timeout = request_timeout
        self.data_getter = data_getter or self.default_data_getter
        self.meta_getter = meta_getter or self.default_meta_getter
        if hasattr(serializer, 'mime'):
            self.serializer = serializer
        else:
            self.serializer = m_data.serializers[serializer]
        if adapter is None:
            if async:
                adapter = m_async.AsyncAdapter()
            else:
                adapter = m_sync.SyncAdapter()
        self.bind_adapter(adapter)

    def bind_adapter(self, adapter):
        adapter.set_header('accept', self.serializer.mime)
        adapter.set_header('content-type', self.serializer.mime)
        adapter.request_timeout = self.request_timeout
        adapter.ingress_filter = self.ingress_filter
        adapter.serializer = self.serializer
        adapter.auth = self.auth
        self.adapter = adapter

    def ingress_filter(self, response):
        """ Flatten a response with meta and data keys into an object. """
        data = self.data_getter(response)
        if isinstance(data, dict):
            data = m_data.DictResponse(data)
        elif isinstance(data, list):
            data = m_data.ListResponse(data)
        else:
            return data
        data.meta = self.meta_getter(response)
        return data

    def do(self, method, path, urn=None, callback=None, data=None, **query):
        urlparts = [self.uri, self.urn if urn is None else urn]
        urlparts.extend(path)
        url = '/'.join(filter(None, (x.strip('/') for x in urlparts)))
        if self.trailing_slash:
            parts = self.urlpartition.split(url, 1)
            if not parts[0].endswith('/'):
                parts[0] += '/'
                url = ''.join(parts)
        return self.adapter.request(method, url, callback=callback, data=data,
                                    query=query)

    def get(self, *path, **query):
        return self.do('get', path, **query)

    def get_pager(self, *path, **query):
        """ A generator for all the results a resource can provide. The pages
        are lazily loaded. """
        fn = self.get_pager_async if self.async else self.get_pager_sync
        page_arg = query.pop('page_size', None)
        limit_arg = query.pop('limit', None)
        query['limit'] = page_arg or limit_arg or self.default_page_size
        return fn(path=path, query=query)

    def get_pager_sync(self, path=None, query=None):
        page = self.get(*path, **query)
        for x in page:
            yield x
        while page.meta['next']:
            page = self.get(urn=page.meta['next'])
            for x in page:
                yield x

    def get_pager_async(self, path=None, query=None):
        return AsyncPager(getter=self.get, path=path, query=query)

    def post(self, *path_and_data, **query):
        path = list(path_and_data)
        data = path.pop(-1)
        return self.do('post', path, data=data, **query)

    def delete(self, *path, **query):
        data = query.pop('data', None)
        return self.do('delete', path, data=data, **query)

    def put(self, *path_and_data, **query):
        path = list(path_and_data)
        data = path.pop(-1)
        return self.do('put', path, data=data, **query)

    def patch(self, *path_and_data, **query):
        path = list(path_and_data)
        data = path.pop(-1)
        return self.do('patch', path, data=data, **query)


class AsyncPager(object):

    max_overflow = 1000

    def __init__(self, getter=None, path=None, query=None):
        self.mark = 0
        self.active = None
        self.waiting = collections.deque()
        self.getter = getter
        self.path = path
        self.query = query
        self.stop = False
        self.next_page = None

    def __iter__(self):
        return self

    def __next__(self):
        item = concurrent.Future()
        self.queue_next(item)
        return item

    next = __next__

    def queue_next_page(self):
        if self.next_page:
            self.active = self.getter(urn=self.next_page)
        else:
            self.active = self.getter(*self.path, **self.query)
        self.active.add_done_callback(self.on_next_page)

    def queue_next(self, item):
        if len(self.waiting) >= self.max_overflow:
            raise OverflowError('max overflow exceeded')
        if self.active:
            if self.active.done():
                if self.active.result():
                    item.set_result(self.active.result().pop(0))
                elif self.stop:
                    raise StopIteration()
                else:
                    self.waiting.append(item)
                    self.queue_next_page()
            else:
                self.waiting.append(item)
        else:
            self.waiting.append(item)
            self.queue_next_page()

    def on_next_page(self, page):
        res = page.result()
        self.next_page = res.meta['next']
        self.stop = not self.next_page
        while self.waiting and res:
            self.waiting.popleft().set_result(res.pop(0))
        if self.waiting:
            if self.stop:
                while self.waiting:
                    self.waiting.popleft().set_exception(StopIteration())
            else:
                self.queue_next_page()


class Resource(dict):
    """ Extended dictionary format that makes future operations on this object
    more object-mapper like. """

    def do(self, method, *path, **query):
        """ XXX: Debatable existence. """
        raise NotImplementedError()

    def fetch(self, method, *path, **query):
        """ Get a subresource. """
        raise NotImplementedError()

    def save(self):
        raise NotImplementedError()

    def delete(self, method, *path, **query):
        raise NotImplementedError()



"""
Interface(s) for adapters subclasses.  Eventually common authentication
should find its way here.
"""

from __future__ import print_function, division

import collections


Response = collections.namedtuple('Response', ('http_code', 'headers',
                                  'content', 'error', 'extra'))


class AdapterBase(object):
    """ Adapter interface.  Must subclass. """

    def set_header(self, header, value):
        """ Set a header that will be included in every HTTP request. """
        raise NotImplementedError('pure virtual method')

    def request(self, method, url, data=None, callback=None, query=None):
        raise NotImplementedError('pure virtual method')



'''
Serialize data to/from foreign data types into python.
'''

from __future__ import print_function

import collections
import datetime
import dateutil.parser
import json
import re
from xml.etree import ElementTree

Serializer = collections.namedtuple('Serializer', 'mime, encode, decode')


class DictResponse(dict):
    pass


class ListResponse(list):
    pass


class NormalJSONEncoder(json.JSONEncoder):
    """ Normal == iso datatime. """

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        else:
            return super(NormalJSONEncoder, self).default(obj)


class NormalJSONDecoder(json.JSONDecoder):
    """ Normal == iso datatime. """

    strict_iso_match = re.compile('\d{4}-\d{2}-\d{2}(T\d{2}:\d{2}\:\d{2}.*)?$')

    def __init__(self):
        super(NormalJSONDecoder, self).__init__(object_hook=self.parse_object)

    def parse_object(self, data):
        """ Look for datetime looking strings. """
        for key, value in data.items():
            if isinstance(value, (str, type(u''))) and \
               self.strict_iso_match.match(value):
                data[key] = dateutil.parser.parse(value)
        return data

serializers = {
    'json': Serializer('application/json',
                       NormalJSONEncoder().encode,
                       NormalJSONDecoder().decode),
    'xml': Serializer('text/xml',
                      ElementTree.tostring,
                      ElementTree.fromstring)
}



#!/usr/bin/env python

from setuptools import setup, find_packages

README = 'README.md'

def long_desc():
    try:
        import pypandoc
    except ImportError:
        with open(README) as f:
            return f.read()
    else:
        return pypandoc.convert(README, 'rst')

setup(
    name='syndicate',
    version='1.1.0',
    description='A wrapper for REST APIs',
    author='Justin Mayfield',
    author_email='tooker@gmail.com',
    url='https://github.com/mayfield/syndicate/',
    license='MIT',
    long_description=long_desc(),
    packages=find_packages(),
    install_requires=[
        'requests',
        'python-dateutil',
        'tornado',
    ],
    test_suite='test',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries',
    ]
)



"""
Asyncronous adapter using the 'tornado' engine.
"""

from __future__ import print_function, division

import functools
import json
import sys
from syndicate.adapters import base
from tornado import httpclient, concurrent, gen
try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode


class AsyncAdapter(base.AdapterBase):

    def __init__(self, *args, **kwargs):
        config = kwargs.pop('config', {})
        self.client = httpclient.AsyncHTTPClient(**config)
        self.headers = {}
        self.request_timeout = None
        self.connect_timeout = None
        super(AsyncAdapter, self).__init__(*args, **kwargs)

    def set_header(self, header, value):
        self.headers[header] = value

    def request(self, method, url, data=None, query=None, callback=None,
                timeout=None):
        user_result = concurrent.TracebackFuture()
        if callback is not None:
            user_result.add_done_callback(callback)
        if data is not None:
            data = self.serializer.encode(data)
        if query:
            url = '%s?%s' % (url, urlencode(query, doseq=True))
        timeout = self.request_timeout if timeout is None else timeout
        request = httpclient.HTTPRequest(url, method=method.upper(),
                                         body=data, headers=self.headers,
                                         request_timeout=timeout,
                                         connect_timeout=self.connect_timeout)
        start = lambda f: self.start_request(f, request, user_result)
        self.authenticate(request).add_done_callback(start)
        return user_result

    def authenticate(self, request):
        f = None
        if callable(self.auth):
            f = self.auth(request)
        elif self.auth is not None:
            request.auth_username, request.auth_password = self.auth
        if f is None:
            # TODO: Replace with gen.maybe_future in tornado 3.3
            f = concurrent.Future()
            f.set_result(None)
        return f

    def start_request(self, auth_result, request, user_result):
        if auth_result.exception():
            concurrent.chain_future(auth_result, user_result)
            return
        try:
            f = self.client.fetch(request)
        except Exception:
            user_result.set_exc_info(sys.exc_info())
        else:
            cb = functools.partial(self.on_request_done, user_result)
            f.add_done_callback(cb)

    def on_request_done(self, user_result, fetch_result):
        """ Finally parse the result and run the user's callback. """
        if fetch_result.exception():
            concurrent.chain_future(fetch_result, user_result)
        else:
            native_resp = fetch_result.result()
            content = None
            try:
                if native_resp.body and len(native_resp.body):
                    content = self.serializer.decode(native_resp.body.decode())
            except Exception:
                user_result.set_exc_info(sys.exc_info())
            else:
                resp = base.Response(http_code=native_resp.code,
                                     headers=native_resp.headers,
                                     content=content, error=None,
                                     extra=native_resp)
                user_result.set_result(self.ingress_filter(resp))


class LoginAuth(object):
    """ Auth where you need to perform an arbitrary "login" to get a cookie.
    The expectation is that the args to this constructor can be used to
    perform a request that generates the required cookie(s) for a valid
    session. """

    content_type = 'application/json'
    OK_HTTP_CODES = 200, 201

    def __init__(self, url, method='POST', **kwargs):
        headers = {
            'content-type': self.content_type
        }
        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
        kwargs['headers'] = headers
        if 'data' in kwargs:
            kwargs['body'] = self.serializer(kwargs.pop('data'))
        self.url = url
        self.method = method
        self.req_kwargs = kwargs
        self.login = None

    @gen.coroutine
    def __call__(self, request):
        if not self.login:
            http = httpclient.AsyncHTTPClient()
            self.login = http.fetch(self.url, method=self.method.upper(),
                                    **self.req_kwargs)
        response = yield self.login
        self.check_login_response()
        request.headers['cookie'] = response.headers['set-cookie']

    def check_login_response(self):
        if self.login.result().code not in self.OK_HTTP_CODES:
            raise Exception('login failed')

    def serializer(self, data):
        return json.dumps(data)


class HeaderAuth(object):
    """ A simple header based auth.  Instantiate this with header name and
    value arguments for a single header or with a dictionary of name/value
    pairs. """

    def __init__(self, header_or_headers_dict, value=None):
        if hasattr(header_or_headers_dict, 'items'):
            self.headers = header_or_headers_dict.copy()
        else:
            self.headers = {header_or_headers_dict: value}

    def __call__(self, request):
        request.headers.update(self.headers)



"""
Syncronous adapter based on the 'requests' library.
"""

from __future__ import print_function, division

import json
import requests
from syndicate.adapters import base


class SyncAdapter(base.AdapterBase):

    def __init__(self, *args, **kwargs):
        config = kwargs.pop('config', {})
        self.session = requests.Session(**config)
        self.request_timeout = None
        super(SyncAdapter, self).__init__(*args, **kwargs)

    def set_header(self, header, value):
        self.session.headers[header] = value

    @property
    def auth(self):
        return self.session.auth

    @auth.setter
    def auth(self, value):
        self.session.auth = value

    def request(self, method, url, data=None, query=None, callback=None,
                timeout=None):
        if data is not None:
            data = self.serializer.encode(data)
        timeout = self.request_timeout if timeout is None else timeout
        resp = self.session.request(method, url, data=data, params=query,
                                    timeout=timeout)
        content = None
        try:
            if resp.content and len(resp.content):
                content = self.serializer.decode(resp.content.decode())
        except Exception as e:
            error = e
            content = None
        else:
            error = None
        r = base.Response(http_code=resp.status_code, headers=resp.headers,
                          content=content, error=error, extra=resp)
        data = self.ingress_filter(r)
        if callback:
            callback(data)
        return data


class HeaderAuth(requests.auth.AuthBase):
    """ A simple header based auth.  Instantiate this with header name and
    value arguments for a single header or with a dictionary of name/value
    pairs. """

    def __init__(self, header_or_headers_dict, value=None):
        if hasattr(header_or_headers_dict, 'items'):
            self.headers = header_or_headers_dict.copy()
        else:
            self.headers = {header_or_headers_dict: value}

    def __call__(self, request):
        request.headers.update(self.headers)
        return request


class LoginAuth(requests.auth.AuthBase):
    """ Auth where you need to perform an arbitrary "login" to get a cookie.
    The expectation is that the args to this constructor can be used to
    perform a request that generates the required cookie(s) for a valid
    session. """

    content_type = 'application/json'
    OK_HTTP_CODES = 200, 201

    def __init__(self, url, method='POST', **kwargs):
        headers = {
            'content-type': self.content_type
        }
        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
        kwargs['headers'] = headers
        if 'data' in kwargs:
            kwargs['data'] = self.serializer(kwargs['data'])
        self.url = url
        self.method = method
        self.req_kwargs = kwargs
        self.login = None

    def __call__(self, request):
        if not self.login:
            self.login = requests.request(self.method, self.url,
                                          **self.req_kwargs)
        self.check_login_response()
        request.prepare_cookies(self.login.cookies)
        return request

    def check_login_response(self):
        if self.login.status_code not in self.OK_HTTP_CODES:
            raise Exception('login failed')

    def serializer(self, data):
        return json.dumps(data)



"""
Sanity tests for the syndicate library.
"""

from __future__ import print_function, division

import datetime
import syndicate
import syndicate.adapters.async
import syndicate.adapters.sync
import syndicate.data
import unittest


class BaseTest(unittest.TestCase):

    def notest_service_construct(self):
        self.assertRaises(TypeError, syndicate.Service)
        syndicate.Service(uri='foo', urn='bar')

    def test_json_serializer(self):
        input_ = {
            "duck": {
                "talk": "Quack!",
                "walk": 0.001,
                "is_duck": True,
                "born": datetime.datetime.utcnow()
            }
        }
        data = syndicate.data.serializers['json'].encode(input_)
        output = syndicate.data.serializers['json'].decode(data)
        self.assertEqual(input_, output)


class AuthTests(unittest.TestCase):

    sync = syndicate.adapters.sync
    async = syndicate.adapters.async

    def request(self):
        class Request(object): pass
        r = Request()
        r.headers = {}
        return r

    def test_header_auth_single(self):
        for Auth in (self.sync.HeaderAuth, self.async.HeaderAuth):
            request = self.request()
            auth = Auth("foo", "bar")
            auth(request)
            self.assertEqual(request.headers, {'foo': 'bar'})
            auth = Auth("bar", "foo")
            auth(request)
            self.assertEqual(request.headers, {'foo': 'bar', 'bar': 'foo'})

    def test_header_auth_multi(self):
        for Auth in (self.sync.HeaderAuth, self.async.HeaderAuth):
            request = self.request()
            auth = Auth({"foo": "bar"})
            auth(request)
            self.assertEqual(request.headers, {'foo': 'bar'})
            auth = Auth({"bar": "foo"})
            auth(request)
            self.assertEqual(request.headers, {'foo': 'bar', 'bar': 'foo'})
            auth = Auth({"foo": "replace"})
            auth(request)
            self.assertEqual(request.headers, {'foo': 'replace', 'bar': 'foo'})


class URLTests(unittest.TestCase):

    valid_path_signatures = {
        (): '%s',
        ('',): '%s',
        ('', ''): '%s',
        ('one', ''): '%s/one',
        ('', 'one'): '%s/one',
        ('', 'one', ''): '%s/one',
        ('/', 'one', ''): '%s/one',
        ('/one', '', ''): '%s/one',
        ('/one',): '%s/one',
        ('/one/',): '%s/one',
        ('one/',): '%s/one',
        ('one',): '%s/one',
        ('one', '/two'): '%s/one/two',
        ('one', '/two/'): '%s/one/two',
        ('one', 'two/'): '%s/one/two',
        ('one/', '/two'): '%s/one/two',
        ('/one/', '/two/'): '%s/one/two',
        # ('//foo//one/', '/two/'): '%s/one/two',  # TODO support this.
    }

    def snoop_request(self, service, snoop_callback):
        service.adapter.request = lambda m, url, **na: snoop_callback(url)

    def clean_url_filter(self, url):
        scheme, sep, uri = url.partition('://')
        self.assertEqual(scheme, 'https')
        self.assertEqual(sep, '://')  # ?
        self.assertNotIn('//', uri)
        return uri

    def test_uri_and_urn_slash_collapse(self):
        uri = 'tld/urn'
        for tld in ('tld', 'tld/'):
            for urn in ('urn', 'urn/', '/urn', '/urn/'):
                s = syndicate.Service(uri='https://%s' % tld, urn=urn)
                self.snoop_request(s, self.clean_url_filter)
                for args, valid_fmt in self.valid_path_signatures.items():
                    self.assertEqual(s.get(*args), (valid_fmt % uri) + '/')
                s = syndicate.Service(uri='https://%s' % tld, urn=urn, trailing_slash=False)
                self.snoop_request(s, self.clean_url_filter)
                for args, valid_fmt in self.valid_path_signatures.items():
                    self.assertEqual(s.get(*args), valid_fmt % uri)

    def test_no_urn(self):
        uri = 'tld'
        for tld in ('tld', 'tld/'):
            s = syndicate.Service(uri='https://%s' % tld)
            self.snoop_request(s, self.clean_url_filter)
            for args, valid_fmt in self.valid_path_signatures.items():
                self.assertEqual(s.get(*args), (valid_fmt % uri) + '/')
            s = syndicate.Service(uri='https://%s' % tld, trailing_slash=False)
            self.snoop_request(s, self.clean_url_filter)
            for args, valid_fmt in self.valid_path_signatures.items():
                self.assertEqual(s.get(*args), valid_fmt % uri)

    def test_slashes_with_hidden_query(self):
        s = syndicate.Service(uri='https://tld/')
        self.snoop_request(s, self.clean_url_filter)
        self.assertEqual(s.get('foo?hide=me'), 'tld/foo/?hide=me')
        self.assertEqual(s.get(urn='foo?hide=me'), 'tld/foo/?hide=me')
